//can animate everything that has props: x,y,alpha,width, height, tint
Lib.animFx = (function() {

    var obj = {};
    var common = {
        opts: undefined,
        hasCallback: undefined,
        obj: undefined,
        callback: undefined,
        context: undefined
    }

    var Common = function(opts) {
        this.allRequired = true;
        this.obj = undefined;
        this.hasCallback = false;
        this.callback = undefined;
        this.context = undefined;

        if (_checkObjParam(opts.obj)) {
            this.obj = opts.obj;
        }
        else {
            //constructor returns empty obj instead of this.
            return {};
        }

        if (_checkCallback(opts.callback, opts.context)) {
            this.callback = opts.callback;
            this.context = opts.context;
            this.hasCallback = true;
        }
    }

    //#########################################################################
    //PRIVATE------------------------------------------------------------
    //#########################################################################
    function _checkObjParam(obj) {
        var objReqProps = ["x", "y", "alpha", "width", "height", "tint", "angle"];
        // p(obj);
        // p(!Lib.check.objHasProps(obj,objReqProps));

        if (!Lib.check.isObject(obj) || !Lib.check.objHasProps(obj, objReqProps)) {
            // opts.obj = Lib.UTILS.createDummyObj(objReqProps); 
            Lib.glob.warn("Obj parameter missing (sprite,image or text). Returning -1.");
            return false;
        }

        return true;
    }

    function _checkCallback(callback, context) {
        if (!Lib.check.isFunction(callback) || !Lib.check.isObject(context)) {
            // Lib.glob.warn
            return false;
        }
        return true;
    }

    function _checkAxisType(val) {
        var validStrings = ["x", "y", "xy"];
        if (Lib.check.arrHasVal(validStrings, val))
            return true;
        else
            return false;
    }



    //returns a single tween with multiple childes based on array
    function _createCombinedTweensMovement(obj, offsetArray, origX = 0, origY = 0, offsetMultiplier = 1, inverseSpeed = 1) {
        //arg  check
        if (!_checkObjParam(obj) || !Lib.check.totalArrCheck(offsetArray, "number")) {
            Lib.glob.warn("Can't create chained tweens, obj or the array are invalid");
            return -1;
        }

        var tween = game.add.tween(obj);
        var easing = Phaser.Easing.Linear.Out;
        var speedDecreaseRatio = 1; //decreases speed for further tweens, otherwise same boring speed

        var pointA = new Phaser.Point(obj.x, obj.y);
        var pointB = new Phaser.Point(origX, origY);

        var targetPointsArray = _getArrayOfPointsBasedOnOffset(pointA, pointB, offsetArray, offsetMultiplier);

        for (var a = 0; a < targetPointsArray.length; a++) {

            var targetX = targetPointsArray[a].x;
            var targetY = targetPointsArray[a].y;

            tween.to({
                x: targetX,
                y: targetY
            }, Lib.DEFS.SECOND * inverseSpeed * speedDecreaseRatio, easing);


            //set it to 0.6 after first big tween
            if (speedDecreaseRatio > 0.6)
                speedDecreaseRatio = 0.6;

            //keep decreasing it gradually until 0.1
            speedDecreaseRatio -= 0.1;
            if (speedDecreaseRatio < 0.1)
                speedDecreaseRatio = 0.1;
        }

        return tween;
    }

    //gets an array of points along the line between A and B based on offset
    function _getArrayOfPointsBasedOnOffset(pointA, pointB, offsetArray, offsetMultiplier) {
        var arr = [];

        var distance = Lib.UTILS.distanceBetweenAB(pointA, pointB);
        for (var a = 0; a < offsetArray.length; a++) {
            var offset = offsetArray[a];
            var newDistance = distance + offset * offsetMultiplier;

            var point = Lib.UTILS.pointTowardsB(pointA, pointB, newDistance);
            arr.push(point);
        }

        return arr;
    }



    function _createCombinedTweensScale(obj, ratioArray, inverseSpeed = 1) {
        //arg check
        if (!_checkObjParam(obj) || !Lib.check.totalArrCheck(ratioArray, "object")) {
            Lib.glob.warn("Can't create chained tweens. Obj ar the array are invalid");
            return -1;
        }
        for (var i = 0; i < ratioArray.length; i++) {
            var ratio = ratioArray[i];
            if (!Lib.check.objHasOwnProps(ratio, ["w", "h"])) {
                Lib.glob.warn("Object in array doesn't have w or h props");
                return -1;
            }
        }

        var tween = game.add.tween(obj);
        var easing = Phaser.Easing.Linear.Out;
        var speedDecreaseRatio = 1;

        for (var a = 0; a < ratioArray.length; a++) {
            var ratio = ratioArray[a];
            var ratioX = obj.width * ratio.w;
            p(ratioX);
            tween.to({
                width: obj.width * ratio.w,
                height: obj.height * ratio.h
            }, Lib.DEFS.SECOND * inverseSpeed * speedDecreaseRatio, easing);


            //set it to 0.4 after first big tween
            if (speedDecreaseRatio > 0.8)
                speedDecreaseRatio = 0.8;

            //keep decreasing it gradually until 0.1
            speedDecreaseRatio -= 0.1;
            if (speedDecreaseRatio < 0.2)
                speedDecreaseRatio = 0.2;
        }

        return tween;

    }

    //#########################################################################
    //PUBLIC--------------------------------------------------------------------
    //######################################################################### 

    function travelTo(opts) {
        //arg check --------------------------------------------
        var opts = opts || {};
        var common = new Common(opts);
        if (Lib.check.isEmptyOrUndefinedObj(common)) {
            return -1;
        }

        var type = Lib.check.isNumber(opts.type) ? opts.type : Lib.DEFS.ANIM_FX.TRAVEL_TO.LINEAR;
        var x = Lib.check.isNumber(opts.x) ? opts.x : 0;
        var y = Lib.check.isNumber(opts.y) ? opts.y : 0;
        var inverseSpeed = Lib.check.isNumber(opts.inverseSpeed) ? opts.inverseSpeed : 1;
        var axis = Lib.check.isString(opts.axis) && !Lib.check.isEmptyString(opts.axis) && _checkAxisType(opts.axis) ? opts.axis : "x";
        var alphaFade = Lib.check.isBoolean(opts.alphaFade) ? opts.alphaFade : true;
        var inertiaTweenData = Lib.check.totalArrCheck(opts.inertiaTweenData,"number") ? opts.inertiaTweenData : Lib.data.tw.inertiaStop;
        var offsetMultiplier = Lib.check.isNumber(opts.offsetMultiplier) ? opts.offsetMultiplier : 6;
        //------------------------------------------------------

        var startPoint = common.obj.position;
        var targetPoint = new Phaser.Point(x, y);

        // _pointsAlongALine(startPoint, targetPoint);

        // p(targetPoint);
        // p(startPoint);

        //***************************************************
        //helper funcs----------------------------------------
        //***************************************************

        function _linear() {
            var tween = game.add.tween(common.obj);
            tween.to({
                x: x,
                y: y
            }, Lib.DEFS.SECOND * inverseSpeed, Phaser.Easing.Quadratic.Out);

            if (common.hasCallback) {
                tween.onComplete.add(common.callback, common.context);
            }

            tween.start();
        }

        function _inertia() {
            var tween = _createCombinedTweensMovement(common.obj, inertiaTweenData, x, y, offsetMultiplier, inverseSpeed, axis);

            if (common.hasCallback) {
                tween.onComplete.add(common.callback, common.context);
            }

            tween.start();
        }

        //--------------------------------------------------------

        if (type === Lib.DEFS.ANIM_FX.TRAVEL_TO.LINEAR)
            _linear();
        else if (type === Lib.DEFS.ANIM_FX.TRAVEL_TO.INERTIA)
            _inertia();

        //add alpha fade
        if (alphaFade) {
            var alphaTween = game.add.tween(common.obj);
            var alphaMultiplier = 4;
            var alphaTime = Lib.DEFS.SECOND * inverseSpeed * alphaMultiplier;
            alphaTween.from({
                alpha: 0
            }, alphaTime, Phaser.Easing.Cubic.Out, true);
        }
    }

    obj.travelTo = travelTo;

    //#########################################################################
    //jellyVibe
    function jellyVibe(opts) {
        //arg check--------------------------
        var opts = opts || {};
        var common = new Common(opts);
        if (Lib.check.isEmptyOrUndefinedObj(common)) {
            return -1;
        }
        var inverseSpeed = Lib.check.isNumber(opts.inverseSpeed) ? opts.inverseSpeed : 0.2;
        //-------------------------------------------------------

        var tween = _createCombinedTweensScale(common.obj, Lib.data.tw.jelly, inverseSpeed);

        if (common.hasCallback) {
            tween.onComplete.add(common.callback, common.context);
        }

        tween.start();
    }

    obj.jellyVibe = jellyVibe;

    //#########################################################################
    //recommended easings: Elastic.Out, Bounce.Out, Exponential.InOut
    //warning: passing a type (Lib.DEFS.ANIM_FX.ROTATE.BOUNCE for example) will ignore passed degrees,inverse speed and easing
    function rotate(opts) {
        //arg check ---------------------------------------
        var opts = opts || {};
        var common = new Common(opts);
        if (Lib.check.isEmptyOrUndefinedObj(common)) {
            return -1;
        }
        var inverseSpeed = Lib.check.isNumber(opts.inverseSpeed) ? opts.inverseSpeed : 1.5;
        var degrees = Lib.check.isNumber(opts.degrees) ? opts.degrees : 720;
        var easing = Lib.check.isFunction(opts.easing) ? opts.easing : Phaser.Easing.Exponential.InOut;
        var type = Lib.check.isNumber(opts.type) ? opts.type : Lib.DEFS.NONE;
        //-------------------------------------------------------

        var tween = game.add.tween(common.obj);

        if (type === Lib.DEFS.ANIM_FX.ROTATE.BOUNCE) {
            tween.to({
                angle: 360
            }, Lib.DEFS.SECOND * 1.2, Phaser.Easing.Bounce.Out);
        }
        else if (type === Lib.DEFS.ANIM_FX.ROTATE.EXPONENTIAL) {
            tween.to({
                angle: 360
            }, Lib.DEFS.SECOND * 1.2, Phaser.Easing.Exponential.InOut);
        }
        else if (type === Lib.DEFS.ANIM_FX.ROTATE.ELASTIC) {
            tween.to({
                angle: 360
            }, Lib.DEFS.SECOND * 3, Phaser.Easing.Elastic.Out);
        }
        else {
            tween.to({
                angle: degrees
            }, Lib.DEFS.SECOND * inverseSpeed, easing); //ignore type, use degrees, speed and easing provided
        }


        if (common.hasCallback) {
            tween.onComplete.add(common.callback, common.context);
        }
        tween.start();
    }

    obj.rotate = rotate;

    //#########################################################################
    function colorFlash(opts) {
        //arg check -------------------------------------------
        var opts = opts || {};
        var common = new Common(opts);
        if (Lib.check.isEmptyOrUndefinedObj(common)) {
            return -1;
        }

        var inverseSpeed = Lib.check.isNumber(opts.inverseSpeed) ? opts.inverseSpeed : 1;
        var targetCol = Lib.check.isNumber(opts.targetCol) ? opts.targetCol : 0xffffff;
        var originalCol = common.obj.tint;
        //--------------------------------------------------------------------------------

        var tween = game.add.tween(common.obj);

        tween.to({
            tint: targetCol
        }, Lib.DEFS.SECOND * inverseSpeed, Phaser.Easing.Quadratic.In);
        tween.to({
            tint: originalCol
        }, Lib.DEFS.SECOND * inverseSpeed, Phaser.Easing.Quadratic.Out);

        if (common.hasCallback) {
            tween.onComplete.add(common.callback, common.context);
        }

        tween.start();
    }
    obj.colorFlash = colorFlash;
    //#########################################################################

    function alphaTarget(opts) {
        //arch check
        var opts = opts || {};
        var common = new Common(opts);
        if (Lib.check.isEmptyOrUndefinedObj(common)) {
            return -1;
        }
        var inverseSpeed = Lib.check.isNumber(opts.inverseSpeed) ? opts.inverseSpeed : 1;
        var targetAlpha = Lib.check.isNumber(opts.targetAlpha) ? opts.targetAlpha : 1;
        var easing = Lib.check.isFunction(opts.easing) ? opts.easing : Phaser.Easing.Linear.Out;
        //-------------------------------------------------

        var tween = game.add.tween(common.obj);

        tween.to({
            alpha: targetAlpha
        }, Lib.DEFS.SECOND * inverseSpeed, easing);

        if (common.hasCallback) {
            tween.onComplete.add(common.callback, common.context);
        }

        tween.start();
    }
    obj.alphaTarget = alphaTarget;

    //rotates and object back and forth repeatedly while altering the angle and the speed
    function pendulumRotateCont(opts) {
        var opts = opts || {};
        var common = new Common(opts);
        if (Lib.check.isEmptyOrUndefinedObj(common)) {
            return -1;
        }
        //Reccomended - BACK, SINUSOIDAL, LINEAR, QUADRATIC
        var easing = Lib.check.isFunction(opts.easing) ? opts.easing : Phaser.Easing.Sinusoidal.InOut;
        var minAngle = Lib.check.isNumber(opts.minAngle) ? opts.minAngle : 20;
        var maxAngle = Lib.check.isNumber(opts.maxAngle) ? opts.maxAngle : 50;
        //the mininum angle rotation that has to happen. Will override min and max angle
        //implemented because a very short angleOffset will appear jittery and jerky
        //Also increase it if the inverse speed min and Max are small
        var guaranteedAngle = Lib.check.isNumber(opts.guaranteedAngle) ? opts.guaranteedAngle: 10;
        var inverseSpeedMin = Lib.check.isNumber(opts.inverseSpeedMin) ? opts.inverseSpeedMin : 200;
        var inverseSpeedMax = Lib.check.isNumber(opts.inverseSpeedMax) ? opts.inverseSpeedMax : 550;




        var tween = game.add.tween(common.obj);
        var tweenBack = game.add.tween(common.obj);
        var currentAngle = common.obj.angle;
        var angleOffset = undefined;
        var inverseSpeed = undefined;
        var duration = undefined;

        function _rotate() {
            // tween.stop();
            tween = game.add.tween(common.obj);
            // p(tween);
            angleOffset = Lib.UTILS.randFloat(minAngle, maxAngle);
            //guaranteed angle modification. Overrides angleOffset
            if(angleOffset < 0){
                if(angleOffset > -guaranteedAngle){
                    angleOffset = -guaranteedAngle;
                }
            }
            else if(angleOffset > 0){
                if(angleOffset < guaranteedAngle){
                    angleOffset = guaranteedAngle;
                }
            }
            else if(angleOffset === 0){
                angleOffset = guaranteedAngle;
            }


            // p("Angle OFFSET : " + angleOffset);
            // p("Angle Offset " + angleOffset);
            // p("OBject Angle " + common.obj.angle);
            var targetAngle = currentAngle + angleOffset;
            // p("Target Angle " + targetAngle);
            //difference between two angles - the shortest between them
            // var diffAngle = Phaser.Math.getShortestAngle(currentAngle,targetAngle);
             inverseSpeed = Lib.UTILS.randFloat(inverseSpeedMin, inverseSpeedMax);
             // p("Inverse Speed " + inverseSpeed);
             //Make angle offset absolute if negative angles were passed
             duration = Math.abs(angleOffset) * inverseSpeed;


            tween.to({angle: targetAngle}, duration,easing);
            tween.yoyo(true);
            tween.start();
            tween.onComplete.add(_rotate,this);
        }
        _rotate();



    }

    obj.pendulumRotateCont = pendulumRotateCont;

    //create a repeating random animation with rand time and randspeeds
    //if anim doesn't exist or if it has just 1 frame the func will return
    function animRandRepeat(opts){
        //required arg. If not passed function will return
        var anim = Lib.check.isObject(opts.anim) && Lib.check.objHasProps(opts.anim, ["speed"]) && Lib.check.objHasMethod(opts.anim,"play") && opts.anim.frameTotal >1? opts.anim : {};
        if(Lib.check.isEmptyOrUndefinedObj(anim)){
            Lib.glob.warn("NO animation object passed. Function will return -1");
            return -1;
        }
        var minRandTime = Lib.check.isNumber(opts.minRandTime) ? opts.minRandTime : 7;
        var maxRandTime = Lib.check.isNumber(opts.maxRandTime) ? opts.maxRandTime : 13;
        //if no min or max speed is provided the animation speed is used - thus no rand speed
        var minSpeed = Lib.check.isNumber(opts.minSpeed) ? opts.minSpeed : opts.anim.speed;
        var maxSpeed = Lib.check.isNumber(opts.maxSpeed) ? opts.maxSpeed : opts.anim.speed;


        //play anim with rand speed
        function _playAnim(){
            var speed = Lib.UTILS.randFloat(minSpeed,maxSpeed);
            anim.speed = speed;
            anim.play();
        }

        //Trigger anim play at random time
        var randTime = Lib.UTILS.randFloat(minRandTime,maxRandTime);
        var event = game.time.events.add(Phaser.Timer.SECOND * randTime, nextEvent,this);

        function nextEvent(){
             randTime = Lib.UTILS.randFloat(minRandTime,maxRandTime);
             p("Rand time " + randTime);
            _playAnim();
            event = game.time.events.add(Phaser.Timer.SECOND * randTime,nextEvent,this);
        }
    }

    obj.animRandRepeat = animRandRepeat;

    //-------------
    return obj;
    //-----------

}());
